/*
    Copyright (c) 2014 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS485p, imported from libdSvDC.

    dS485p is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS485p is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_AVAHI

#include <assert.h>
#include <errno.h>
#include <poll.h>
#include <signal.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>

#include <avahi-client/client.h>
#include <avahi-client/lookup.h>
#include <avahi-client/publish.h>

#include <avahi-common/alternative.h>
#include <avahi-common/error.h>
#include <avahi-common/domain.h>
#include <avahi-common/llist.h>
#include <avahi-common/malloc.h>
#include <avahi-common/simple-watch.h>
#include <avahi-common/timeval.h>

#include "discovery.h"
#include "vdsmRole.h"
#include "whitelist.h"
#include "base/log.h"
#include "base/util.h"
#include "base/ds_string.h"

static bool discover_ds485d;
static struct ds485_avahi_callbacks g_cb;

static int g_running;
static bool g_whitelist_changed = false;
static pthread_t avahi_tid;

static const char* g_default_avahi_service_name = "digitalSTROM ds485p";
static uint16_t g_ds485p_listen_port = 0;
static char g_ds485p_dsuid_str[DSUID_STR_LEN];
static bool g_ds485p_collecting = true;  // Announce as a collecting ds485p

static const char* g_ds485p_service_type_str = "_ds-ds485p._tcp";
static const char* g_vdsm_service_type_str = "_ds-vdsm._tcp";
static const char* g_485d_service_type_str = "_ds-485d._tcp";

/* Legacy DSS service identification: both the type and the name must match. */
static const char* g_old_dss_service_type_str = "_http._tcp";
static const char* g_old_dss_service_name = "dSS";

typedef enum {
    SERVICE_TYPE_UNKNOWN,
    SERVICE_TYPE_DS485P,
    SERVICE_TYPE_OLD_DSS,
    SERVICE_TYPE_VDSM,
    SERVICE_TYPE_485D,
} service_type_t;

service_type_t service_type_from_string(const char* type_str) {
    if (strcmp(type_str, g_ds485p_service_type_str) == 0) {
        return SERVICE_TYPE_DS485P;
    } else if (strcmp(type_str, g_old_dss_service_type_str) == 0) {
        return SERVICE_TYPE_OLD_DSS;
    } else if (strcmp(type_str, g_vdsm_service_type_str) == 0) {
        return SERVICE_TYPE_VDSM;
    } else if (strcmp(type_str, g_485d_service_type_str) == 0) {
        return SERVICE_TYPE_485D;
    } else {
        return SERVICE_TYPE_UNKNOWN;
    }
}

const char* service_type_string(service_type_t type) {
    switch (type) {
    case SERVICE_TYPE_DS485P:
        return g_ds485p_service_type_str;
    case SERVICE_TYPE_OLD_DSS:
        return g_old_dss_service_type_str;
    case SERVICE_TYPE_VDSM:
        return g_vdsm_service_type_str;
    case SERVICE_TYPE_485D:
        return g_485d_service_type_str;
    default:
        return NULL;
    }
}

typedef struct ds485p_avahi_handle {
    AvahiSimplePoll *simple_poll;
    AvahiClient *client;
    AvahiEntryGroup *group;
    char* name;
    uint16_t service_port;
    char *service_dsuid;
    bool collecting;
} ds485p_avahi_handle_t;

typedef struct ServiceInfo ServiceInfo;
struct ServiceInfo {
    ds485p_avahi_handle_t *handle;
    AvahiIfIndex interface;
    AvahiProtocol protocol;
    char *name;
    service_type_t type;
    char *domain;
    bool resolved;

    AvahiServiceResolver *resolver;
};

typedef struct ServiceInfo485d ServiceInfo485d;
struct ServiceInfo485d {
    ServiceInfo service;
    char* uri;
    dsuid_t dsuid;
    bool autoconnected;
    AVAHI_LLIST_FIELDS(ServiceInfo485d, info);
};

static ServiceInfo485d *services_485d = NULL;

typedef struct ServiceInfoVdsm ServiceInfoVdsm;
struct ServiceInfoVdsm {
    ServiceInfo service;
    char* uri;
    dsuid_t dsuid;
    bool autoconnected;
    VDSMRole role;
    AVAHI_LLIST_FIELDS(ServiceInfoVdsm, info);
};

static ServiceInfoVdsm *services_vdsm = NULL;

typedef struct ServiceInfoDS485p ServiceInfoDS485p;
struct ServiceInfoDS485p {
    ServiceInfo service;
    bool collecting;
    AVAHI_LLIST_FIELDS(ServiceInfoDS485p, info);
};

static ServiceInfoDS485p *services_ds485p = NULL;

// g_collection_enabled == true indicates that forward collection is enabled:
// New vdsm are collected as they are discovered. It is set to false while
// other ds485p(s) are present in the network.
static bool g_collection_enabled = false;

// g_disable_backward_collection is used to disable the backward-collection of
// vdSMs discovered during initialization if another ds485p was also collecting
// at any point during initialization.
static bool g_disable_backward_collection = false;

// Globals to keep track of progress during initialization.
static bool g_initial_ds485p_browse_complete = false;
static bool g_initial_old_dss_browse_complete = false;
static bool g_coordination_resolve_complete = false;  // Initialization complete

static void ds485p_clear_discovery_globals() {
    g_collection_enabled = false;
    g_disable_backward_collection = false;
    g_initial_ds485p_browse_complete = false;
    g_initial_old_dss_browse_complete = false;
    g_coordination_resolve_complete = false;
}

bool coordination_browse_complete() {
    return g_initial_ds485p_browse_complete &&
        g_initial_old_dss_browse_complete;
}

static bool compute_device_autoconnect_and_update_whitelist(
    const char *name,
    const dsuid_t *dsuid,
    bool noauto,
    bool collection_enabled)
{
    const bool autoconnect_disabled = noauto || !collection_enabled;
    DSUID_STR(dsuid_str);
    dsuid_to_string(dsuid, dsuid_str);

    if (whitelist_exists()) {
        vdsm_whitelist_fields_t fields;
        fields.noauto = noauto;
        fields.autoconnect = !autoconnect_disabled;
        fields.name = name;
        const bool autoconnect =
            whitelist_insert_and_get_autoconnect(dsuid, &fields);
        if (!autoconnect) {
            report(
                LOG_INFO,
                "Not connecting device %s because it is not whitelisted as autoconnect.\n",
                dsuid_str);
            return false;
        } else {
            report(
                LOG_DEBUG,
                "device %s whitelisted as autoconnect. Connecting...\n",
                dsuid_str);
        }
    } else {
        if (noauto) {
            report(
                LOG_INFO,
                "not connecting NOAUTO device %s because there is no whitelist\n",
                dsuid_str);
            return false;
        } else {
            report(
                LOG_DEBUG,
                "No whitelist, and device %s has NOAUTO == false. Connecting...\n",
                dsuid_str);
        }
    }
    return true;
}

static bool try_connect_to_device(
    const char *uri,
    const char *name,
    const dsuid_t *dsuid,
    bool noauto,
    bool collection_enabled)
{
    const bool autoconnect =
        compute_device_autoconnect_and_update_whitelist(
            name, dsuid, noauto, collection_enabled);
    if (autoconnect) {
        g_cb.bus_discovered(uri, name, dsuid);
    }
    return autoconnect;
}

static bool try_connect_to_vdsm(ServiceInfoVdsm *si_vdsm,
                                bool collection_enabled) {
    if (!si_vdsm->role.collectable) {
        report(LOG_DEBUG, "vdSM is not collectable: \"%s\"\n",
               si_vdsm->service.name);
        return false;
    }
    if (si_vdsm->role.auxiliary) {
        bool master_present = false;
        ServiceInfoVdsm *i = NULL;
        for (i = services_vdsm; i; i = i->info_next) {
            if (i->role.master) {
                master_present = true;
                break;
            }
        }
        if (master_present) {
            report(LOG_DEBUG,
                   "Ignoring auxiliary vdSM \"%s\": master present.\n");
            return false;
        }
    }
    return try_connect_to_device(
        si_vdsm->uri,
        si_vdsm->service.name,
        &si_vdsm->dsuid,
        si_vdsm->role.noauto,
        collection_enabled);
}

static void vdsm_collection_initialized(bool collection_enabled) {
    ServiceInfoVdsm *si_vdsm;
    for (si_vdsm = services_vdsm; si_vdsm; si_vdsm = si_vdsm->info_next) {
        if (!si_vdsm->service.resolved || si_vdsm->uri == NULL) {
            continue;
        }
        report(LOG_INFO,
               "Considering postponed vdsm: %s\n",
               si_vdsm->service.name);
        si_vdsm->autoconnected = try_connect_to_vdsm(
            si_vdsm, collection_enabled);
    }
    ServiceInfo485d *si_485d;
    for (si_485d = services_485d; si_485d; si_485d = si_485d->info_next) {
        if (!si_485d->service.resolved || si_485d->uri == NULL) {
            continue;
        }
        report(LOG_INFO,
               "Considering postponed 485d: %s\n",
               si_485d->service.name);
        si_485d->autoconnected = try_connect_to_device(
            si_485d->uri,
            si_485d->service.name,
            &si_485d->dsuid,
            false, /* noauto */
            collection_enabled);
    }
}

static void ds485p_whitelist_changed_callback() {
    if (!g_coordination_resolve_complete) {
        return;
    }
    report(LOG_DEBUG,
           "Whitelist changed. Reprocessing discovered devices.\n");
    ServiceInfoVdsm *si_vdsm;
    for (si_vdsm = services_vdsm; si_vdsm; si_vdsm = si_vdsm->info_next) {
        if (!si_vdsm->service.resolved || si_vdsm->uri == NULL) {
            continue;
        }
        if (si_vdsm->autoconnected) {
            const bool autoconnect = whitelist_get_autoconnect(&si_vdsm->dsuid);
            if (!autoconnect) {
                DSUID_STR(dsuid_str);
                dsuid_to_string(&si_vdsm->dsuid, dsuid_str);
                report(
                    LOG_DEBUG,
                    "Autoconnect not set for %s in the whitelist.\n",
                    dsuid_str);
                si_vdsm->autoconnected = false;
                g_cb.bus_left(si_vdsm->service.name);
            }
        } else {
            si_vdsm->autoconnected =
                try_connect_to_vdsm(si_vdsm, g_collection_enabled);
        }
    }
    ServiceInfo485d *si_485d;
    for (si_485d = services_485d; si_485d; si_485d = si_485d->info_next) {
        if (!si_485d->service.resolved || si_485d->uri == NULL) {
            continue;
        }
        if (si_485d->autoconnected) {
            const bool autoconnect = whitelist_get_autoconnect(&si_485d->dsuid);
            if (!autoconnect) {
                DSUID_STR(dsuid_str);
                dsuid_to_string(&si_485d->dsuid, dsuid_str);
                report(
                    LOG_DEBUG,
                    "Autoconnect not set for %s in the whitelist.\n",
                    dsuid_str);
                si_485d->autoconnected = false;
                g_cb.bus_left(si_485d->service.name);
            }
        } else {
            si_485d->autoconnected =
                try_connect_to_device(
                    si_485d->uri,
                    si_485d->service.name,
                    &si_485d->dsuid,
                    false, /* noauto */
                    g_collection_enabled);
        }
    }
}

static bool all_foreign_ds485ps_resolved()
{
    // Checks for both ds485p and old dSS.
    bool found_unresolved_foreign_ds485p = false;
    ServiceInfoDS485p *si_ds485p = NULL;
    for (si_ds485p = services_ds485p; si_ds485p; si_ds485p = si_ds485p->info_next) {
        if (!si_ds485p->service.resolved) {
            found_unresolved_foreign_ds485p = true;
            break;
        }
    }
    return !found_unresolved_foreign_ds485p;
}

static void update_coordination_state()
{
    if (!coordination_browse_complete() || !all_foreign_ds485ps_resolved()) {
        return;
    }
    /* All the initial ds485p resolver requests have been completed */

    bool initial_resolver_completion = false;
    if (!g_coordination_resolve_complete) {
        g_coordination_resolve_complete = true;
        /* This the first time we see all ds485p services resolved */
        initial_resolver_completion = true;
        report(LOG_INFO,
               "Collection initialization complete with %s ds485p conflict(s).\n",
               g_disable_backward_collection ? "some" : "no" );
        vdsm_collection_initialized(!g_disable_backward_collection);
    }

    bool found_foreign_collecting_ds485p = false;
    ServiceInfoDS485p *si_ds485p = NULL;
    for (si_ds485p = services_ds485p; si_ds485p; si_ds485p = si_ds485p->info_next) {
        if (si_ds485p->collecting) {
            found_foreign_collecting_ds485p = true;
            break;
        }
    }

    const bool collection_previously_enabled = g_collection_enabled;
    g_collection_enabled = !found_foreign_collecting_ds485p;

    if (initial_resolver_completion ||
        g_collection_enabled != collection_previously_enabled) {
        const char* state = (g_collection_enabled) ? "ENABLED" : "RESTRICTED";
        report(LOG_INFO, "vdSM collection: %s\n", state);
    }
}

static void create_resolver(ServiceInfo *info);

static bool service_matches(
    ServiceInfo* i,
    AvahiIfIndex interface,
    AvahiProtocol protocol,
    const char *name,
    service_type_t type,
    const char *domain) {
    return (i->interface == interface &&
            i->protocol == protocol &&
            strcasecmp(i->name, name) == 0 &&
            i->type == type &&
            avahi_domain_equal(i->domain, domain));
}

static ServiceInfo485d *find_service_485d(
    AvahiIfIndex interface,
    AvahiProtocol protocol,
    const char *name,
    service_type_t type,
    const char *domain)
{
    ServiceInfo485d *i;
    for (i = services_485d; i; i = i->info_next) {
        if (service_matches(&i->service, interface, protocol, name, type, domain)) {
            return i;
        }
    }
    return NULL;
}

static ServiceInfoVdsm *find_service_vdsm(
    AvahiIfIndex interface,
    AvahiProtocol protocol,
    const char *name,
    service_type_t type,
    const char *domain)
{
    ServiceInfoVdsm *i;
    for (i = services_vdsm; i; i = i->info_next) {
        if (service_matches(&i->service, interface, protocol, name, type, domain)) {
            return i;
        }
    }
    return NULL;
}

static ServiceInfoDS485p *find_service_ds485p(
    AvahiIfIndex interface,
    AvahiProtocol protocol,
    const char *name,
    service_type_t type,
    const char *domain)
{
    ServiceInfoDS485p *i;
    for (i = services_ds485p; i; i = i->info_next) {
        if (service_matches(&i->service, interface, protocol, name, type, domain)) {
            return i;
        }
    }
    return NULL;
}

static ServiceInfo *add_service(ds485p_avahi_handle_t *handle,
                                AvahiIfIndex interface,
                                AvahiProtocol protocol,
                                const char *name,
                                service_type_t type,
                                const char *domain)
{
    ServiceInfo *i;
    if (type == SERVICE_TYPE_485D) {
        if (find_service_485d(interface, protocol, name, type, domain)) {
            return NULL;
        }
        ServiceInfo485d *si_485d = avahi_new(struct ServiceInfo485d, 1);
        if (!si_485d) {
            panic("OOM: service info new");
        }
        si_485d->uri = NULL;
        si_485d->autoconnected = false;
        AVAHI_LLIST_PREPEND(ServiceInfo485d, info, services_485d, si_485d);
        i = &si_485d->service;
    } else if (type == SERVICE_TYPE_VDSM) {
        if (find_service_vdsm(interface, protocol, name, type, domain)) {
            return NULL;
        }
        ServiceInfoVdsm *si_vdsm = avahi_new(struct ServiceInfoVdsm, 1);
        if (!si_vdsm) {
            panic("OOM: service info new");
        }
        si_vdsm->uri = NULL;
        si_vdsm->autoconnected = false;
        vdsm_role_clear(&si_vdsm->role);
        AVAHI_LLIST_PREPEND(ServiceInfoVdsm, info, services_vdsm, si_vdsm);
        i = &si_vdsm->service;
    } else if (type == SERVICE_TYPE_DS485P ||
               type == SERVICE_TYPE_OLD_DSS) {
        bool is_self = (type == SERVICE_TYPE_DS485P &&
                        strcmp(handle->name, name) == 0);
        if (is_self) {
            return NULL;
        }
        if (find_service_ds485p(interface, protocol, name, type, domain)) {
            return NULL;
        }
        ServiceInfoDS485p *si_ds485p = avahi_new(struct ServiceInfoDS485p, 1);
        if (!si_ds485p) {
            panic("OOM: service info new");
        }
        if (type == SERVICE_TYPE_OLD_DSS) {
            si_ds485p->collecting = true;
            g_disable_backward_collection = true;
        } else {
            si_ds485p->collecting = false;  // Set in the resolver callback.
        }
        AVAHI_LLIST_PREPEND(ServiceInfoDS485p, info, services_ds485p, si_ds485p);
        i = &si_ds485p->service;
    } else {
        return NULL;
    }

    i->handle = handle;
    i->interface = interface;
    i->protocol = protocol;
    i->name = avahi_strdup(name);
    i->type = type;
    i->domain = avahi_strdup(domain);
    i->resolved = false;

    return i;
}

static void clear_service(ServiceInfo* i) {
    assert(i);

    if (i->resolver) {
        avahi_service_resolver_free(i->resolver);
    }

    avahi_free(i->name);
    avahi_free(i->domain);
}

static void remove_service_485d(ServiceInfo485d* i) {
    AVAHI_LLIST_REMOVE(ServiceInfo485d, info, services_485d, i);
    clear_service(&i->service);
    avahi_free(i->uri);
    avahi_free(i);
}

static void remove_service_vdsm(ServiceInfoVdsm* i) {
    AVAHI_LLIST_REMOVE(ServiceInfoVdsm, info, services_vdsm, i);
    clear_service(&i->service);
    avahi_free(i->uri);
    avahi_free(i);
}

static void remove_service_ds485p(ServiceInfoDS485p* i) {
    AVAHI_LLIST_REMOVE(ServiceInfoDS485p, info, services_ds485p, i);
    clear_service(&i->service);
    avahi_free(i);
}

/* Returns true if the service removal signals a bus removal */
static bool remove_service(
    AvahiIfIndex interface,
    AvahiProtocol protocol,
    const char *name,
    service_type_t type,
    const char *domain) {

    if (type == SERVICE_TYPE_485D) {
        ServiceInfo485d *si_485d =
            find_service_485d(interface, protocol, name, type, domain);
        if (!si_485d) {
            return false;
        }
        remove_service_485d(si_485d);
        return true;
    } else if (type == SERVICE_TYPE_VDSM) {
        ServiceInfoVdsm *si_vdsm =
            find_service_vdsm(interface, protocol, name, type, domain);
        if (!si_vdsm) {
            return false;
        }
        remove_service_vdsm(si_vdsm);
        return true;
    } else if (type == SERVICE_TYPE_DS485P ||
               type == SERVICE_TYPE_OLD_DSS) {
        ServiceInfoDS485p *si_ds485p =
            find_service_ds485p(interface, protocol, name, type, domain);
        if (!si_ds485p) {
            return false;
        }
        remove_service_ds485p(si_ds485p);
        update_coordination_state();
        return false;
    }
    return false;
}

bool parse_vdsm_role_txt_records(AvahiStringList *txt, VDSMRole* role)
{
    role->collectable = (avahi_string_list_find(txt, "collectable") != NULL);
    role->collecting = (avahi_string_list_find(txt, "collecting") != NULL);
    role->master = (avahi_string_list_find(txt, "master") != NULL);
    role->auxiliary = (avahi_string_list_find(txt, "auxiliary") != NULL);
    role->noauto = (avahi_string_list_find(txt, "noauto") != NULL);
    return vdsm_role_sanitize(role);
}

static void ds485p_avahi_resolve_callback(
        AvahiServiceResolver *resolver,
        AVAHI_GCC_UNUSED AvahiIfIndex interface,
        AVAHI_GCC_UNUSED AvahiProtocol protocol,
        AvahiResolverEvent event,
        const char *name,
        const char *type_str,
        const char *domain,
        const char *host_name,
        const AvahiAddress *address,
        uint16_t port,
        AvahiStringList *txt,
        AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
        void* userdata)
{
    assert(resolver);
    assert(userdata);

    ServiceInfo *info = userdata;
    ds485p_avahi_handle_t *handle = info->handle;

    /* decouple info->resolver in case of retry */
    assert(info->resolver == resolver);
    info->resolver = NULL;

    /*
     * callback is called whenever a service has been
     * resolved successfully or timed out
     */
    switch (event) {
    case AVAHI_RESOLVER_FAILURE:
        if (avahi_client_errno(handle->client) == AVAHI_ERR_TIMEOUT) {
            /*
             * service consists of PTR/TXT/SRV DNS
             * descriptors, if one of them has shorter
             * time-to-live and expires earlier, resolving
             * fails. Keep retrying until success or
             * resolving fails with an error other than
             * timeout e.g. service has expired from cache
             * completely
             */
            report(LOG_NOTICE,
                   "resolver timeout:: service:\"%s\" type:\"%s\" domain:\"%s\"\n",
                   name, type_str, domain);
            create_resolver(info);
            break;
        }

        report(LOG_ERR,
               "resolver failure:: service:\"%s\" type:\"%s\" domain:\"%s\": err:%s\n",
               name, type_str, domain,
               avahi_strerror(avahi_client_errno(handle->client)));
        avahi_simple_poll_quit(handle->simple_poll);
        break;

    case AVAHI_RESOLVER_FOUND: {
        char ip[AVAHI_ADDRESS_STR_MAX];
        avahi_address_snprint(ip, sizeof(ip), address);
        report(LOG_NOTICE, "found %s at: %s:%u (%s)\n", name, host_name,
               port, ip);

        const service_type_t type = service_type_from_string(type_str);
        if (type == SERVICE_TYPE_DS485P ||
            type == SERVICE_TYPE_OLD_DSS) {
            ServiceInfoDS485p* si_ds485p =
                find_service_ds485p(interface, protocol, name, type, domain);
            if (si_ds485p) {
                si_ds485p->service.resolved = true;
                if (type == SERVICE_TYPE_DS485P) {
                    si_ds485p->collecting =
                        (avahi_string_list_find(txt, "collecting") != NULL);
                    if (si_ds485p->collecting) {
                        g_disable_backward_collection = true;
                    }
                }
                update_coordination_state();
            }
        } else {
            char *dsuid_str = NULL;
            AvahiStringList *pair = avahi_string_list_find(txt, "dSUID");
            if (pair) {
                dsuid_str = strchr((const char *)avahi_string_list_get_text(pair),
                                   '=');
                if (dsuid_str) {
                    // skip '='
                    dsuid_str++;
                }
            }
            if (!dsuid_str) {
                report(LOG_DEBUG,
                       "ignoring vdSM %s at: %s:%u (%s) - missing dSUID in announcement\n",
                       name, host_name, port, ip);
                break;
            }
            size_t buflen = /* tcp:// */ 6 + strlen(host_name) + /* : */ 1 +
                /* 65535 */ 5 + /* NULL */ 1;
            char *uri = malloc(buflen);
            if (!uri) {
                report(LOG_ERR, "could not allocate memory for bus uri "
                       "(%s:%u / %s)\n", host_name, port, name);
                return;
            }
            snprintf(uri, buflen, "tcp://%s:%u", host_name, port);

            dsuid_t dsuid;
            if (dsuid_from_string(dsuid_str, &dsuid) != 0) {
                report(LOG_ERR, "failed to convert string to dsuid\n");
                return;
            }

            if (type == SERVICE_TYPE_VDSM) {
                ServiceInfoVdsm* si_vdsm =
                    find_service_vdsm(interface, protocol, name, type, domain);
                if (si_vdsm) {
                    if (parse_vdsm_role_txt_records(txt, &si_vdsm->role)) {
                        si_vdsm->service.resolved = true;
                        si_vdsm->uri = uri;
                        si_vdsm->dsuid = dsuid;
                        uri = NULL;
                        if (g_coordination_resolve_complete) {
                            si_vdsm->autoconnected =
                                try_connect_to_vdsm(si_vdsm, g_collection_enabled);
                        }
                    } else {
                        report(LOG_ERR,
                               "Ignorigh vdSM with invalid txt records: %s\n",
                               dsuid_str);
                    }
                }
            } else if (type == SERVICE_TYPE_485D) {
                ServiceInfo485d* si_485d =
                    find_service_485d(interface, protocol, name, type, domain);
                if (si_485d) {
                    si_485d->service.resolved = true;
                    si_485d->uri = uri;
                    si_485d->dsuid = dsuid;
                    uri = NULL;
                    if (g_coordination_resolve_complete) {
                        si_485d->autoconnected =
                            try_connect_to_device(si_485d->uri, name, &dsuid,
                                                  false, /* noauto */
                                                  g_collection_enabled);
                    }
                }
            }
            if (uri) {
                free(uri);
            }
        }
        break;
    }
    default:
        break;
    }

    avahi_service_resolver_free(resolver);
}

static void create_resolver(ServiceInfo *info)
{
    ds485p_avahi_handle_t *handle = info->handle;
    AvahiServiceResolver *resolver;
    /*
     * We ignore the returned resolver object. In the callback
     * function we free it. If the server is terminated before
     * the callback function is called the server will free
     * the resolver for us.
     */
    resolver = avahi_service_resolver_new(handle->client, info->interface,
                                          info->protocol, info->name,
                                          service_type_string(info->type),
                                          info->domain, AVAHI_PROTO_INET, 0,
                                          ds485p_avahi_resolve_callback, info);
    if (!resolver) {
        if (avahi_client_errno(handle->client) == AVAHI_ERR_NO_MEMORY) {
            panic("OOM: create resolver");
        }

        report(LOG_ERR, "create resolver failed service:\"%s\": err:%s\n ",
               info->name, avahi_strerror(avahi_client_errno(handle->client)));
        /* start over */
        avahi_simple_poll_quit(handle->simple_poll);
    }

    /*
     * could the callback have fired before, we get here?
     * we are running on the mainloop, so this is likely save
     */
    info->resolver = resolver;
}

static void ds485p_avahi_browse_callback(
        AvahiServiceBrowser *browser,
        AvahiIfIndex interface,
        AvahiProtocol protocol,
        AvahiBrowserEvent event,
        const char *name,
        const char *type_str,
        const char *domain,
        AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
        void* userdata)
{
    assert(browser);
    assert(userdata);

    ds485p_avahi_handle_t *avahi = userdata;
    const service_type_t type = service_type_from_string(type_str);

    /* called when a new services becomes available or is removed from LAN */

    switch (event) {
    case AVAHI_BROWSER_FAILURE:

        report(LOG_ERR, "avahi browse callback %s\n",
               avahi_strerror(avahi_client_errno(avahi_service_browser_get_client(browser))));
        avahi_simple_poll_quit(avahi->simple_poll);
        return;

    case AVAHI_BROWSER_NEW: {
        if (type == SERVICE_TYPE_OLD_DSS &&
            strcmp(name, g_old_dss_service_name) != 0) {
            return;
        }
        ServiceInfo *info;
        report(LOG_DEBUG,
               "browser: new service \"%s\" of type \"%s\" in domain \"%s\"\n",
               name, type_str, domain);
        info = add_service(avahi, interface, protocol, name, type, domain);
        if (info) {
            create_resolver(info);
        }
        break;
    }

    case AVAHI_BROWSER_REMOVE: {
        if (type == SERVICE_TYPE_OLD_DSS &&
            strcmp(name, g_old_dss_service_name) != 0) {
            return;
        }
        report(LOG_DEBUG,
               "browser: remove service \"%s\" of type \"%s\" in domain '%s'\n",
               name, type_str, domain);

        if (remove_service(interface, protocol, name, type, domain)) {
            g_cb.bus_left(name);
        }
        break;
    }

    case AVAHI_BROWSER_ALL_FOR_NOW: {
        report(LOG_DEBUG, "browser: ALL_FOR_NOW\n");
        if (type == SERVICE_TYPE_DS485P || type == SERVICE_TYPE_OLD_DSS) {
            if (type == SERVICE_TYPE_DS485P) {
                g_initial_ds485p_browse_complete = true;
            }
            if (type == SERVICE_TYPE_OLD_DSS) {
                g_initial_old_dss_browse_complete = true;
            }
            if (coordination_browse_complete()) {
                update_coordination_state();
            }
        }
        break;
    }

    case AVAHI_BROWSER_CACHE_EXHAUSTED:
        report(LOG_DEBUG, "browser: CACHE_EXHAUSTED\n");
    default:
        break;
    }
}

static bool ds485p_avahi_create_services(ds485p_avahi_handle_t *handle);

static bool ds485p_avahi_service_name_collision(ds485p_avahi_handle_t *handle)
{
    if (!handle || !handle->name || !handle->group) {
        report(LOG_ERR,
               "service_name_collision requires a service name and entry group.");
        return false;
    }
    char* new_name = avahi_alternative_service_name(handle->name);
    avahi_free(handle->name);
    handle->name = new_name;
    report(LOG_WARNING, "service name collision, renaming service to '%s'\n",
           handle->name);
    avahi_entry_group_reset(handle->group);
    return ds485p_avahi_create_services(handle);
}

static void ds485p_avahi_entry_group_callback(AvahiEntryGroup *group,
                                              AvahiEntryGroupState state,
                                              void *userdata)
{
    assert(group);
    assert(userdata);

    ds485p_avahi_handle_t *handle = (ds485p_avahi_handle_t *)userdata;
    if (avahi_entry_group_get_client(group) != handle->client) {
        report(LOG_ERR, "entry group callback error: bad client.\n");
        return;
    }
    /*
     * CAUTION: this callback might have fired before
     * userdata->group has been set
     */
    handle->group = group; /* idempotent */

    switch (state) {
    case AVAHI_ENTRY_GROUP_ESTABLISHED:
        report(LOG_DEBUG, "service \"%s\" successfully established.\n",
               handle->name);
        break;
    case AVAHI_ENTRY_GROUP_COLLISION:
        ds485p_avahi_service_name_collision(handle);
        break;
    case AVAHI_ENTRY_GROUP_FAILURE :
        report(LOG_ERR, "entry group failure: %s\n",
            avahi_strerror(avahi_client_errno(handle->client)));
        /* failure while registering service */
        avahi_simple_poll_quit(handle->simple_poll);
        break;
    case AVAHI_ENTRY_GROUP_UNCOMMITED:
    case AVAHI_ENTRY_GROUP_REGISTERING:
    default:
        break;
    }
}

static AvahiStringList* ds485p_avahi_create_txt_records(ds485p_avahi_handle_t *handle)
{
    const char *key = "dSUID=";
    size_t len = strlen(key) + strlen(handle->service_dsuid) + 1;
    char *dsuid = malloc(len);
    if (!dsuid) {
        return NULL;
    }
    snprintf(dsuid, len, "%s%s", key, handle->service_dsuid);
    AvahiStringList* list = avahi_string_list_new(dsuid, NULL);
    if (!list) {
        return NULL;
    }
    if (handle->collecting) {
        list = avahi_string_list_add(list, "collecting");
        if (!list) {
            return NULL;
        }
    }
    return list;
}

static bool ds485p_avahi_create_services(ds485p_avahi_handle_t *handle)
{
    if (!handle->group) {
        handle->group = avahi_entry_group_new(handle->client,
                                              ds485p_avahi_entry_group_callback,
                                              (void*)handle);
        if (!handle->group) {
            report(LOG_ERR, "could not create avahi_entry_group_new: %s\n",
                   avahi_strerror(avahi_client_errno(handle->client)));
            return false;
        }
    }
    if (!avahi_entry_group_is_empty(handle->group)) {
        report(LOG_NOTICE, "avahi_create_services: services already present.\n");
        return true;
    }
    report(LOG_DEBUG, "Adding service '%s'\n", handle->name);

    AvahiStringList *txt = ds485p_avahi_create_txt_records(handle);
    if (!txt) {
        report(LOG_ERR, "could not create _ds-ds485p._tcp txt records.\n");
        return false;
    }

    int ret = avahi_entry_group_add_service_strlst(
        handle->group,
        AVAHI_IF_UNSPEC,
        AVAHI_PROTO_INET, 0,
        handle->name,
        g_ds485p_service_type_str,
        NULL,                     /* domain: let avahi decide. */
        NULL,                     /* host: let avahi decide. */
        handle->service_port,
        txt);

    avahi_string_list_free(txt);

    if (ret < 0) {
        if (ret == AVAHI_ERR_COLLISION) {
            return ds485p_avahi_service_name_collision(handle);
        }
        report(LOG_ERR, "could not add _ds-ds485p._tcp service: %s\n",
               avahi_strerror(ret));
        return false;
    }

    ret = avahi_entry_group_commit(handle->group);
    if (ret < 0) {
        report(LOG_ERR, "could not commit entry group: %s\n",
               avahi_strerror(ret));
        return false;
    }
    return true;
}

static AvahiServiceBrowser *add_avahi_browser(ds485p_avahi_handle_t *handle,
                                              service_type_t type)
{
    assert(handle->client);
    const char* type_str = service_type_string(type);
    AvahiServiceBrowser* browser =
        avahi_service_browser_new(handle->client, AVAHI_IF_UNSPEC,
                                  AVAHI_PROTO_UNSPEC,
                                  type_str,
                                  NULL,
                                  0,
                                  ds485p_avahi_browse_callback,
                                  handle);
    if (!browser) {
        report(LOG_ERR,
               "could not create avahi browser for service %s: %s\n",
               type_str,
               avahi_strerror(avahi_client_errno(handle->client)));
    }
    return browser;
}

static bool ds485p_avahi_running_client_init(ds485p_avahi_handle_t *handle)
{
    assert(handle->client);
    report(LOG_NOTICE, "avahi-client: running\n");
    if (!ds485p_avahi_create_services(handle)) {
        return false;
    }
    if (!add_avahi_browser(handle, SERVICE_TYPE_VDSM)) {
        return false;
    }
    if (!add_avahi_browser(handle, SERVICE_TYPE_DS485P)) {
        return false;
    }
    if (!add_avahi_browser(handle, SERVICE_TYPE_OLD_DSS)) {
        return false;
    }
    if (discover_ds485d) {
        if (!add_avahi_browser(handle, SERVICE_TYPE_485D)) {
            return false;
        }
    }
    return true;
}


static void ds485p_avahi_client_callback(
        AvahiClient *client,
        AvahiClientState state,
        void *d)
{
    assert(client);
    assert(d);

    /*
     * CAUTION: this callback might have fired before
     * userdata->client has been set
     */
    ds485p_avahi_handle_t *data = d;
    data->client = client; /* idempotent */

    switch (state) {
    case AVAHI_CLIENT_FAILURE:
        report(LOG_ERR, "avahi client: conection failure: %s\n",
               avahi_strerror(avahi_client_errno(client)));
        avahi_simple_poll_quit(data->simple_poll);
        break;

    case AVAHI_CLIENT_S_COLLISION:
        /* Let's drop our registered services. When the server is back
         * in AVAHI_SERVER_RUNNING state we will register them
         * again with the new host name. */
    case AVAHI_CLIENT_S_REGISTERING:
        /* The server records are now being established. This
         * might be caused by a host name change. We need to wait
         * for our own records to register until the host name is
         * properly established. */
        if (data->group) {
            avahi_entry_group_reset(data->group);
        }
        break;
    case AVAHI_CLIENT_S_RUNNING:
        if (!ds485p_avahi_running_client_init(data)) {
            avahi_simple_poll_quit(data->simple_poll);
        }
        break;

    case AVAHI_CLIENT_CONNECTING:
        /*
         * We're still connecting. This state is only
         * entered when AVAHI_CLIENT_NO_FAIL has been passed
         * to avahi_client_new() and the daemon is not yet
         * available.
         */
        report(LOG_NOTICE, "avahi-client: connecting / server not available\n");
        break;
    }
}

static AvahiClient *create_client(ds485p_avahi_handle_t *data)
{
    AvahiClient *client;
    int error;

    /* client callback might fire before function completes */
    client = avahi_client_new(avahi_simple_poll_get(data->simple_poll),
                              AVAHI_CLIENT_NO_FAIL,
                              ds485p_avahi_client_callback,
                              data, &error);
    if (!client) {
        if (error == AVAHI_ERR_NO_MEMORY) {
            panic("OOM: avahi client create");
        } else if (error == AVAHI_ERR_DBUS_ERROR) {
            report(LOG_ERR, "avahi create client:: dbus err:%s\n",
                   avahi_strerror(error));
            return NULL;
        }
        /* shouldn't happen due to AVAHI_CLIENT_NO_FAIL */
        report(LOG_ERR, "avahi create client:: err: %s\n",
               avahi_strerror(error));
        return NULL;
    }

    return client;
}

/* We use ds485p_avahi_poll to override the poll function used by Avahi.
 * This allows to trigger callback(s) in the avahi thread as a response to
 * signals. To handle a signal in this way,
 * 1. Its signal handler should set a global boolean flag.
 * 2. The signals should be blocked during normal execution (in ds485p.c).
 * 3. The signal should be unblocked by ppoll.
 * 4. After ppoll returns, the global flag can be tested and reset.
 *
 * The blocking and unblocking of the signal is needed to ensure that ppoll
 * wakes even if the signal arrives outside of the ppoll call.
 */
int ds485p_avahi_poll(struct pollfd *ufds,
                      unsigned int nfds,
                      int timeout,
                      void *userdata __attribute__((unused))) {
    struct timespec timeout_ts;
    struct timespec *timeout_arg = NULL;
    // timeout unit is milliseconds.
    if (timeout >= 0) {
        timeout_ts.tv_sec = timeout / 1000;
        timeout_ts.tv_nsec = (timeout % 1000) * 1000 * 1000;
        timeout_arg = &timeout_ts;
    }

    sigset_t signals_blocked_during_ppoll;
    int sigmask_res = pthread_sigmask(
        SIG_SETMASK,  // ignored
        NULL,         // No change. Just read the current mask.
        &signals_blocked_during_ppoll);
    if (sigmask_res != 0) {
        report(LOG_ERR, "pthread_sigmask failed: %s", strerror(errno));
    }

    sigdelset(&signals_blocked_during_ppoll, SIGHUP);  // Unblock SIGHUP

    int res;
    int ppoll_errno = 0;
    while (1) {
        res = ppoll(ufds, nfds, timeout_arg, &signals_blocked_during_ppoll);
        if (res == -1) {
            ppoll_errno = errno;
        }

        if (g_whitelist_changed) {
            g_whitelist_changed = false;
            ds485p_whitelist_changed_callback();
            if (res == -1 && ppoll_errno == EINTR) {
                // Returning -1 / EINTR would causes the avahi poll loop to exit.
                continue;
            }
        }
        break;
    }
    if (res == -1) {
        errno = ppoll_errno;
    }

    return res;
}

void ds485p_avahi_shutdown(ds485p_avahi_handle_t *avahi);

static bool ds485p_avahi_init(ds485p_avahi_handle_t *avahi, const char* service_name) {
    avahi->simple_poll = NULL;
    avahi->client = NULL;
    avahi->name = avahi_strdup(service_name);
    avahi->service_port = g_ds485p_listen_port;
    avahi->service_dsuid = g_ds485p_dsuid_str;
    avahi->collecting = g_ds485p_collecting;
    avahi->simple_poll = avahi_simple_poll_new();
    if (!avahi->simple_poll) {
        report(LOG_ERR, "could not allocate avahi_poll object\n");
        return false;
    }
    avahi_simple_poll_set_func(avahi->simple_poll,
                               ds485p_avahi_poll,
                               NULL);

    avahi->client = create_client(avahi);
    if (!avahi->client) {
        /* already logged */
        ds485p_avahi_shutdown(avahi);
        return false;
    }

    return true;
}

void ds485p_avahi_shutdown(ds485p_avahi_handle_t *avahi)
{
    while (services_vdsm) {
        /* has pointer to resolvers, delete before client */
        remove_service_vdsm(services_vdsm);
    }
    while (services_485d) {
        /* has pointer to resolvers, delete before client */
        remove_service_485d(services_485d);
    }
    while (services_ds485p) {
        /* has pointer to resolvers, delete before client */
        remove_service_ds485p(services_ds485p);
    }
    if (avahi->client) {
        /* will free entry groups, browsers and resolvers as well */
        avahi_client_free(avahi->client);
        avahi->group = NULL;
        avahi->client = NULL;
    }
    if (avahi->simple_poll) {
        avahi_simple_poll_free(avahi->simple_poll);
        avahi->simple_poll = NULL;
    }
    if (avahi->name) {
        avahi_free(avahi->name);
        avahi->name = NULL;
    }
}

static void* sysAvahiThread(void *arg)
{
    ds485p_avahi_handle_t *avahi_handle = arg;

    char* service_name = avahi_strdup(g_default_avahi_service_name);
    while (g_running) {
        ds485p_clear_discovery_globals();
        if (ds485p_avahi_init(avahi_handle, service_name)) {
            avahi_free(service_name);
            avahi_simple_poll_loop(avahi_handle->simple_poll);
            service_name = avahi_strdup(avahi_handle->name);
            ds485p_avahi_shutdown(avahi_handle);
        }

        if (!g_running) {
          break;
        }

        report(LOG_WARNING,
               "avahi daemon interrupted or stopped... reconnect.\n");

        /*
         * If ds485p_avahi_init fails we will retry again and again.
         * To avoid a busy loop, we will wait a second.
         */
        sleep(1);
    }
    avahi_free(service_name);

    report(LOG_ERR, "exit avahi discovery\n");
    return NULL;
}

/* don't use these global other than start/stop discover */
static ds485p_avahi_handle_t g_XXX_avahi_handle;

int start_discovery(int filter,
                    struct ds485_avahi_callbacks *cb,
                    const dsuid_t* service_dsuid,
                    uint16_t service_port,
                    bool collecting)
{
    int ret;
    g_running = true;
    g_ds485p_listen_port = service_port;
    g_ds485p_collecting = collecting;

    if (dsuid_to_string(service_dsuid, g_ds485p_dsuid_str) != DSUID_RC_OK) {
        return -EINVAL;
    }

    discover_ds485d = (filter & DISCOVER_DS485D);

    if (cb == NULL) {
        return -EINVAL;
    }
    memcpy(&g_cb, cb, sizeof(*cb));

    ret = pthread_create(&avahi_tid, NULL, sysAvahiThread, &g_XXX_avahi_handle);
    if (ret != 0) {
        report(LOG_ERR, "Could not create avahi thread.\n");
        return ret < 0 ? ret : -ret; /* make sure it's negative */
    }
    pthread_detach(avahi_tid);
    return 0;
}

void stop_discovery()
{
    g_running = false;
    avahi_simple_poll_quit(g_XXX_avahi_handle.simple_poll);
}

void autoconnect_whitelist_changed() {
    // This call is triggered by a signal handler. The signal will wake the
    // avahi loop.
    g_whitelist_changed = true;
}

#endif//HAVE_AVAHI
